import abvio.incar as In
import abvio.io as Io

import unittest
import os
from pathlib import Path

from pymatgen.core import Structure
from pymatgen.io.vasp import Incar


base_path = Path(__file__).parent
structure_dir = os.path.join(base_path, 'structures')
files_dir = os.path.join(base_path, 'files')
vaspset_dir = os.path.join(base_path, 'vaspset')

perovskite_structure = Structure.from_file(os.path.join(structure_dir, 'CaTiO3.vasp'))
fluorite_structure = Structure.from_file(os.path.join(structure_dir, 'CaF2.vasp'))

class TestMagmom(unittest.TestCase):

    def test_assign_magnetic_moments_by_species(self):
        structure = perovskite_structure.copy()
        structure = In.assign_site_property_by_species(structure, 'Ti', 'magmom', 2.0)
        structure = In.assign_site_property_by_species(structure, 'O', 'magmom', 0.0)
        structure = In.assign_site_property_by_species(structure, 'Ca', 'magmom', 0.0)

        for site in structure.sites:
            if site.species_string == 'Ti':
                self.assertEqual(site.properties['magmom'], 2.0)
            else:
                self.assertEqual(site.properties['magmom'], 0.0)

    def test_assign_magnetic_moments_by_species_fluorite(self):
        structure = fluorite_structure.copy()
        structure = In.assign_site_property_by_species(structure, 'F', 'magmom', 1.0)
        structure = In.assign_site_property_by_species(structure, 'Ca', 'magmom', 2.0)
        for site in structure.sites:
            if site.species_string == 'Ca':
                self.assertEqual(site.properties['magmom'], 2.0)
            else:
                self.assertEqual(site.properties['magmom'], 1.0)
        
    def test_assign_magnetic_moments_by_index(self):
        structure = perovskite_structure.copy()

        sites_for_zero = [0, 2 ]
        sites_for_two = [1, 3]
        sites_for_one = [ 4 ]

        for site in sites_for_zero:
            structure = In.assign_site_property_by_index(structure, site, 'magmom', 0.0)
        for site in sites_for_two:
            structure = In.assign_site_property_by_index(structure, site, 'magmom', 2.0)
        for site in sites_for_one:
            structure = In.assign_site_property_by_index(structure, site, 'magmom', 1.0)

        for i, site in enumerate(structure.sites):
            if i in sites_for_zero:
                self.assertEqual(site.properties['magmom'], 0.0)
            elif i in sites_for_two:
                self.assertEqual(site.properties['magmom'], 2.0)
            else:
                self.assertEqual(site.properties['magmom'], 1.0)

    def test_assign_magnetic_moments_by_range(self):

        #fluorite structure has 12 sites (0-11)
        structure = fluorite_structure.copy()

        #every other site has a magmom of 2.0 while the rest have 0.0
        structure = In.assign_site_property_by_range(structure, slice(0, 12, 2), 'magmom', 2.0)
        structure = In.assign_site_property_by_range(structure, slice(1, 12, 2), 'magmom', 1.0)

        #check if magmoms alternate between 2.0 and 0.0
        for i, site in enumerate(structure.sites):
            if i % 2 == 0:
                self.assertEqual(site.properties['magmom'], 2.0)
            else:
                self.assertEqual(site.properties['magmom'], 1.0)


    def test_magnetic_moments_from_structure(self):
        structure = perovskite_structure.copy()
        structure = In.assign_site_property_by_species(structure, 'Ti', 'magmom',2.0)
        structure = In.assign_site_property_by_species(structure, 'O', 'magmom', 0.0)
        structure = In.assign_site_property_by_species(structure, 'Ca', 'magmom', 0.0)

        magmoms = In.site_properties_from_structure(structure, 'magmom')

        self.assertEqual(magmoms, [0.0, 2.0, 0.0, 0.0, 0.0])


class TestNoncollinearMagMom(unittest.TestCase):


    def test_assign_magnetic_moments_by_species(self):
        structure = perovskite_structure.copy()
        structure = In.assign_site_property_by_species(structure, 'Ti', 'magmom', [0, 0, 2.0])
        structure = In.assign_site_property_by_species(structure, 'O', 'magmom', [0, 0, 0])
        structure = In.assign_site_property_by_species(structure, 'Ca', 'magmom', [0, 0, 0])

        for site in structure.sites:
            if site.species_string == 'Ti':
                self.assertEqual(site.properties['magmom'], [0, 0, 2.0])
            else:
                self.assertEqual(site.properties['magmom'], [0, 0, 0])

    def test_magnetic_moments_from_structure(self):
        structure = fluorite_structure.copy()
        structure = In.assign_site_property_by_species(structure, 'Ca', 'magmom', [2.0, 0, -1])

        magmoms = In.site_properties_from_structure(structure, 'magmom', default=[0, 0, 0])
        print(structure)
        self.assertEqual(magmoms, [[2.0, 0, -1]] * 4 + [[0, 0, 0]] * 8)


    def test_assign_magnetic_moments_by_range(self):

        #fluorite structure has 12 sites (0-11)
        structure = fluorite_structure.copy()

        #every other site has a magmom of 2.0 while the rest have 0.0
        structure = In.assign_site_property_by_range(structure, slice(0, 12, 2), 'magmom', [2.0] * 3)
        structure = In.assign_site_property_by_range(structure, slice(1, 12, 2), 'magmom', [0.0] * 3)

        #check if magmoms alternate between 2.0 and 0.0
        for i, site in enumerate(structure.sites):
            if i % 2 == 0:
                self.assertEqual(site.properties['magmom'], [2.0] * 3)
            else:
                self.assertEqual(site.properties['magmom'], [0.0] * 3)


class TestLDAUUFunctions(unittest.TestCase):

    def test_assign_ldauu_by_species(self):
        structure = perovskite_structure.copy()
        structure = In.assign_site_property_by_species(structure, 'Ti', 'ldauu', 2.0)
        structure = In.assign_site_property_by_species(structure, 'O', 'ldauu', 0.0)
        structure = In.assign_site_property_by_species(structure, 'Ca', 'ldauu', 0.0)

        for site in structure.sites:
            if site.species_string == 'Ti':
                self.assertEqual(site.properties['ldauu'], 2.0)
            else:
                self.assertEqual(site.properties['ldauu'], 0.0)

    def test_assign_ldauu_by_index(self):
        structure = perovskite_structure.copy()

        sites_for_zero = [0, 2 ]
        sites_for_two = [1, 3]
        sites_for_one = [ 4 ]

        for site in sites_for_zero:
            structure = In.assign_site_property_by_index(structure, site, 'ldauu', 0.0)
        for site in sites_for_two:
            structure = In.assign_site_property_by_index(structure, site, 'ldauu', 2.0)
        for site in sites_for_one:
            structure = In.assign_site_property_by_index(structure, site, 'ldauu', 1.0)

        for i, site in enumerate(structure.sites):
            if i in sites_for_zero:
                self.assertEqual(site.properties['ldauu'], 0.0)
            elif i in sites_for_two:
                self.assertEqual(site.properties['ldauu'], 2.0)
            else:
                self.assertEqual(site.properties['ldauu'], 1.0)

    def test_assign_ldauu_by_range(self):

        #fluorite structure has 12 sites (0-11)
        structure = fluorite_structure.copy()

        #every other site has a ldauu of 2.0 while the rest have 0.0
        structure = In.assign_site_property_by_range(structure, slice(0, 12, 2), 'ldauu', 2.0)
        structure = In.assign_site_property_by_range(structure, slice(1, 12, 2), 'ldauu', 0.0)

        #check if ldauu alternate between 2.0 and 0.0
        for i, site in enumerate(structure.sites):
            if i % 2 == 0:
                self.assertEqual(site.properties['ldauu'], 2.0)
            else:
                self.assertEqual(site.properties['ldauu'], 0.0)


    def test_ldauu_from_structure(self):
        structure = perovskite_structure.copy()
        structure = In.assign_site_property_by_species(structure, 'Ti', 'ldauu', 2.0)
        structure = In.assign_site_property_by_species(structure, 'O', 'ldauu', 0.0)
        structure = In.assign_site_property_by_species(structure, 'Ca', 'ldauu', 0.0)

        ldauu = In.site_properties_from_structure(structure, 'ldauu')

        self.assertEqual(ldauu, [0.0, 2.0, 0.0, 0.0, 0.0])


class TestReadSiteProperties(unittest.TestCase):

    species_yaml_file = os.path.join(files_dir, 'magmom_species.yaml')
    index_yaml_file = os.path.join(files_dir, 'magmom_index.yaml')
    range_yaml_file = os.path.join(files_dir, 'magmom_range.yaml')

    species_test_dict = Io.load(species_yaml_file)['incar']
    index_test_dict = Io.load(index_yaml_file)['incar']
    range_test_dict = Io.load(range_yaml_file)['incar']

    noncollinear_species_yaml_file = os.path.join(files_dir, 'magmom_noncollinear_species.yaml')
    noncollinear_index_yaml_file = os.path.join(files_dir, 'magmom_noncollinear_index.yaml')
    noncollinear_range_yaml_file = os.path.join(files_dir, 'magmom_noncollinear_range.yaml')

    noncollinear_species_test_dict = Io.load(noncollinear_species_yaml_file)['incar']
    noncollinear_index_test_dict = Io.load(noncollinear_index_yaml_file)['incar']
    noncollinear_range_test_dict = Io.load(noncollinear_range_yaml_file)['incar']

    def test_read_magmom_by_species(self):
        magmom_dict = self.species_test_dict.get('magmom')
        self.assertIsInstance(magmom_dict, dict)

        for key, value in magmom_dict.items():
            self.assertIsInstance(key, str)
            self.assertIsInstance(value, (int, float))

    def test_read_magmom_by_index(self):

        magmom_dict = self.index_test_dict.get('magmom')
        self.assertIsInstance(magmom_dict, dict)

        for key, value in magmom_dict.items():
            self.assertIsInstance(key, int)
            self.assertIsInstance(value, (int, float))

    def test_read_magmom_by_range(self):
        magmom_dict_list = self.range_test_dict.get('magmom')

        self.assertIsInstance(magmom_dict_list, list)
        for magmom_dict in magmom_dict_list:
            self.assertIsInstance(magmom_dict, dict)

            for key, value in magmom_dict.items():
                self.assertIsInstance(key, str)
                self.assertIsInstance(value, (int, float))

    def test_read_noncollinear_magmom_by_species(self):
        magmom_dict = self.noncollinear_species_test_dict.get('magmom')
        self.assertIsInstance(magmom_dict, dict)

        for key, value in magmom_dict.items():
            self.assertIsInstance(key, str)
            self.assertIsInstance(value, list)

    def test_read_noncollinear_magmom_by_index(self):
            
            magmom_dict = self.noncollinear_index_test_dict.get('magmom')
            self.assertIsInstance(magmom_dict, dict)
    
            for key, value in magmom_dict.items():
                self.assertIsInstance(key, int)
                self.assertIsInstance(value, list)

    def test_read_noncollinear_magmom_by_range(self):

        magmom_dict_list = self.noncollinear_range_test_dict.get('magmom')

        self.assertIsInstance(magmom_dict_list, list)
        for magmom_dict in magmom_dict_list:
            self.assertIsInstance(magmom_dict, dict)

            for key, value in magmom_dict.items():
                self.assertIsInstance(key, str)
                self.assertIsInstance(value, (int,list))

    def test_is_index_dict(self):

        index_dicts = [self.index_test_dict, self.noncollinear_index_test_dict]
        non_index_dicts = [self.species_test_dict, self.range_test_dict, self.noncollinear_species_test_dict, self.noncollinear_range_test_dict]

        for index_dict in index_dicts:
            magmom_dict = index_dict.get('magmom')
            self.assertTrue(In.is_index_dict(magmom_dict))

        for non_index_dict in non_index_dicts:
            magmom_dict = non_index_dict.get('magmom')
            self.assertFalse(In.is_index_dict(magmom_dict))

    def test_is_species_dict(self):
            
        species_dicts = [self.species_test_dict, self.noncollinear_species_test_dict]
        non_species_dicts = [self.index_test_dict, self.range_test_dict, self.noncollinear_index_test_dict, self.noncollinear_range_test_dict]
    
        for species_dict in species_dicts:
            magmom_dict = species_dict.get('magmom')
            self.assertTrue(In.is_species_dict(magmom_dict))
    
        for non_species_dict in non_species_dicts:
            magmom_dict = non_species_dict.get('magmom')
            self.assertFalse(In.is_species_dict(magmom_dict))

    def test_is_range_list(self):
            
        range_dicts = [self.range_test_dict, self.noncollinear_range_test_dict]
        non_range_dicts = [self.species_test_dict, self.index_test_dict, self.noncollinear_species_test_dict, self.noncollinear_index_test_dict]
    
        for range_dict in range_dicts:
            magmom_dict_list = range_dict.get('magmom')
            self.assertTrue(In.is_range_list(magmom_dict_list))
    
        for non_range_dict in non_range_dicts:
            magmom_dict_list = non_range_dict.get('magmom')
            self.assertFalse(In.is_range_list(magmom_dict_list))


    def test_is_collinear(self):

        collinear_list = [self.species_test_dict, self.index_test_dict, self.range_test_dict]
        noncollinear_list = [self.noncollinear_species_test_dict, self.noncollinear_index_test_dict, self.noncollinear_range_test_dict]

        for collinear_dict in collinear_list:
            magmom_dict = collinear_dict.get('magmom')
            self.assertTrue(In.is_collinear(magmom_dict))

        for noncollinear_dict in noncollinear_list:
            magmom_dict = noncollinear_dict.get('magmom')
            self.assertFalse(In.is_collinear(magmom_dict))


class TestIncarModel(unittest.TestCase):


    def test_valid_model(self):

        structure = perovskite_structure.copy()
        incar_dict = {
            'incar': {
                'magmom': {
                    'Ti': 2.0,
                    'O': 0.0,
                    'Ca': 0.0
                },
                'ediff': 1e-4,
                'ismear': -5,
                'lreal': False,
                'ediffg': -0.01,
                'lwave': False,
            }
        }

        model = In.IncarModel(incar_dict=incar_dict['incar'])
        self.assertIsInstance(model, In.IncarModel)
        incar = model.incar(structure)
        self.assertIsInstance(incar, Incar)

        incar.check_params()

    def test_valid_noncollinear_model(self):
        structure = perovskite_structure.copy()

        incar_dict = {
            'incar': {
                'magmom': {
                    'Ti': [0, 0, 2.0],
                    'O': [0, 0, 0],
                    'Ca': [0, 0, 0]
                },
                'ediff': 1e-4,
                'ismear': -5,
                'lreal': False,
                'ediffg': -0.01,
                'lwave': False,
            }
        }

        model = In.IncarModel(incar_dict=incar_dict['incar'])
        self.assertIsInstance(model, In.IncarModel)
        incar = model.incar(structure)
        self.assertIsInstance(incar, Incar)

        incar.check_params()
        
        self.assertEqual(incar['MAGMOM'], [[0, 0, 0], [0, 0, 2.0], [0, 0, 0], [0, 0, 0], [0, 0, 0]])

    def test_invalid_magmom_incar_model(self):
        incar_dict = {
            'incar': {
                'magmom': [
                    {'Ti': 2.0},
                    {'O': 0.0},
                    {'Ca': 0.0}
                ],
                'ediff': 1e-4,
                'ismear': -5,
                'lreal': False,
                'ediffg': -0.01,
                'lwave': False,
            }
        }

        with self.assertRaises(ValueError):
            model = In.IncarModel(incar_dict=incar_dict['incar'])        
            model.incar(fluorite_structure)

    def test_valid_model_magmom_range(self):

        structure = fluorite_structure.copy()
        incar_dict = {
            'incar': {
                'magmom': [
                    {'start': 0, 'stop': 12, 'step': 2, 'value': 50.0},
                    {'start': 1, 'stop': 12, 'step': 2, 'value': 7.0}
                ],
                'ediff': 1e-4,
                'ismear': -5,
                'lreal': False,
                'ediffg': -0.01,
                'lwave': False,
            }
        }

        model = In.IncarModel(incar_dict=incar_dict['incar'])
        incar = model.incar(structure)
        incar.check_params()

        self.assertIsInstance(model, In.IncarModel)
        self.assertIsInstance(incar, Incar)

        #check if magmoms alternate between 50.0 and 7.0
        generated_magmom_list = incar['MAGMOM']
        for i, magmom in enumerate(generated_magmom_list):
            if i % 2 == 0:
                self.assertEqual(magmom, 50.0)
            else:
                self.assertEqual(magmom, 7.0)
        



if __name__ == '__main__':
    unittest.main()


import unittest
import os

import numpy as np
import abvio.structure as st
import abvio.io as io

from pymatgen.io.vasp import Poscar
from pymatgen.core import Structure, Lattice
from pathlib import Path

base_path = Path(__file__).parent
structure_dir = os.path.join(base_path, 'structures')
files_dir = os.path.join(base_path, 'files')

def matrix_similarity(matrix1: np.ndarray, matrix2: np.ndarray, tol: float = 1e-6) -> bool:
    """Check if two matrices are similar within a tolerance"""
    return np.allclose(matrix1, matrix2, atol=tol)


class TestFormatingFunctions(unittest.TestCase):
    """Contains tests that check if read can format species correctly"""

    def test_format_species_dict(self):
        """Test if format_species can format a list of dictionaries"""

        species = [{'Fe': 2}, {'O': 4}]
        formatted_species = st.format_species(species)

        self.assertEqual(formatted_species, ['Fe', 'Fe', 'O', 'O', 'O', 'O'])

    
    def test_single_species(self):
        """Test if format_species can format a single species"""

        species = ['Fe']
        formatted_species = st.format_species(species)

        self.assertEqual(formatted_species, ['Fe'])

    def test_format_species_list(self):
        """Test if format_species can format a list of strings"""

        species = ['Fe', 'O']
        formatted_species = st.format_species(species)

        self.assertEqual(formatted_species, ['Fe', 'O'])

    def test_format_species_invalid(self):
        """Test if format_species raises an error for invalid input"""

        invalid_species = ['Fe', 2, {4,0}, {4:0}, [{4:0}]]

        for species in invalid_species:
            with self.assertRaises(ValueError):
                st.format_species(species)


class TestBaseStructure(unittest.TestCase):

    def test_valid_base_structure(self):
        
        st.BaseStructure(mode="external")
        st.BaseStructure(mode="manual")
        st.BaseStructure(mode="prototype")

        st.BaseStructure(mode="external", file="POSCAR")
        st.BaseStructure(mode="external", code="mp-1234")
        st.BaseStructure(mode="external", code="1234")

    def test_invalid_base_structures(self):

        with self.assertRaises(ValueError):

            st.BaseStructure(mode="internal")

        with self.assertRaises(ValueError):
            st.BaseStructure(mode="wrong")

        with self.assertRaises(ValueError):
            st.BaseStructure(mode="JohnCena")


class TestManualStructure(unittest.TestCase):

    def test_valid_manual_structure(self):

        lattice = Lattice.from_parameters(a=1, b=1, c=1, alpha=90, beta=90, gamma=90)
        species = ["Mg", "O"]
        coords = [[0, 0, 0], [0.5, 0.5,0.5]]

        model = st.ManualStructure(lattice=lattice, species=species, coords=coords)
        structure = model.structure

        self.assertIsInstance(structure, Structure)

    def test_valid_manual_structure_from_array(self):

        lattice = [
            [1, 0, 0],
            [0, 1, 0],
            [0, 0, 1]
        ]
        species = ["Mg", "O"]
        coords = [[0, 0, 0], [0.5, 0.5,0.5]]

        model = st.ManualStructure(lattice=lattice, species=species, coords=coords)
        structure = model.structure

        self.assertIsInstance(structure, Structure)

    def test_valid_manual_structure_from_numpy(self):

        lattice = np.array([
            [1, 0, 0],
            [0, 1, 0],
            [0, 0, 1]
        ])
        species = ["Mg", "O"]
        coords = [[0, 0, 0], [0.5, 0.5,0.5]]

        model = st.ManualStructure(lattice=lattice, species=species, coords=coords)
        structure = model.structure

        self.assertIsInstance(structure, Structure)

    def test_invalid_manual_structure(self):

        with self.assertRaises(ValueError):

            lattice = Lattice.from_parameters(a=1, b=1, c=1, alpha=90, beta=90, gamma=120)
            species = ["Mg", "O", "Ti"]
            coords = [[0, 0, 0], [0.5, 0.5,0.5]]

            st.ManualStructure(lattice=lattice, species=species, coords=coords).structure

        with self.assertRaises(ValueError):
                
            lattice = Lattice.from_parameters(a=1, b=1, c=1, alpha=90, beta=40, gamma=90)
            species = ["Mg", "O"]
            coords = [[0, 0, 0], [0.5, 0.5,0.5], [0.5, 0.5,0.5]]

            st.ManualStructure(lattice=lattice, species=species, coords=coords).structure

    def test_manual_structure_from_file(self):
            
            filepath = os.path.join(files_dir, 'lattice_list.yaml')
            input_dict = io.load(filepath)['structure']

            model = st.ManualStructure.validate(input_dict)
            structure = model.structure

            self.assertIsInstance(structure, Structure)


class TestPrototypeStructure(unittest.TestCase):

    perovskite_dict = io.load(os.path.join(files_dir, 'prototype_perovskite.yaml'))['structure']
    fluorite_dict = io.load(os.path.join(files_dir, 'prototype_fluorite.yaml'))['structure']

    pervoskite_structure = Structure.from_file(os.path.join(structure_dir, 'CaTiO3.vasp'))
    fluorite_structure = Structure.from_file(os.path.join(structure_dir, 'CaF2.vasp'))

    def test_valid_pervoskite_structure(self):
        """Compares a known structure with the one generated by the prototype structure"""

        species = self.perovskite_dict['species']
        lattice = self.perovskite_dict['lattice']
        prototype = self.perovskite_dict['prototype']

        model = st.PrototypeStructure(species=species, lattice=lattice, prototype=prototype)
        structure = model.structure

        self.assertIsInstance(structure, Structure)
        self.assertTrue(matrix_similarity(structure.lattice.matrix, self.pervoskite_structure.lattice.matrix))

    def test_valid_fluorite_structure(self):
        """Compares a known structure with the one generated by the prototype structure"""

        species = self.fluorite_dict['species']
        lattice = self.fluorite_dict['lattice']
        prototype = self.fluorite_dict['prototype']

        model = st.PrototypeStructure(species=species, lattice=lattice, prototype=prototype)
        structure = model.structure

        self.assertIsInstance(structure, Structure)
        self.assertTrue(matrix_similarity(structure.lattice.matrix, self.fluorite_structure.lattice.matrix))


class TestStructureFromMaterialProject(unittest.TestCase):

    def test_valid_code(self):
    
        stucture = st.structure_from_mpi_code("mp-5827")
        self.assertIsInstance(stucture, Structure)

class TestExternalStructure(unittest.TestCase):

    def test_model_from_file(self):

        files = [ 'CaTiO3.vasp', 'CaF2.vasp'] 

        for file in files:
            model = st.ExternalStructure(file=os.path.join(structure_dir, file))
            structure = model.structure

            self.assertIsInstance(structure, Structure)


    def test_model_from_string(self):

        string = Poscar.from_file(os.path.join(structure_dir, 'CaTiO3.vasp')).__str__()

        model = st.ExternalStructure(string=string)
        structure = model.structure

        self.assertIsInstance(structure, Structure)


    def test_model_from_materials_project(self):

        model = st.ExternalStructure(code="mp-5827")
        structure = model.structure
        
        self.assertIsInstance(structure, Structure)


class TestStructureFromInputDict(unittest.TestCase):

    def test_valid_input_dict(self):

        input_dict = {
            "mode": "external",
            "file": os.path.join(structure_dir, 'CaTiO3.vasp')
        }

        model = st.structure_model_from_input_dict(input_dict)
        structure = model.structure
        self.assertIsInstance(structure, Structure)

    def test_invalid_input_dict(self):

        input_dict = {
            "mode": "external",
            "file": os.path.join(structure_dir, 'CaTiO3.vasp'),
            "string": "string"
        }

        with self.assertRaises(ValueError):
            st.structure_model_from_input_dict(input_dict)


if __name__ == "__main__":
    unittest.main()


import logging
import warnings
import yaml
import os

from pathlib import Path

from pymatgen.core import Structure
from pymatgen.io.vasp import Incar, Kpoints, Poscar

from abvio.structure import structure_model_from_input_dict
from abvio.kpoints import kpoints_model_from_dictionary
from abvio.incar import IncarModel

log = logging.getLogger(__name__)
log_format = "%(asctime)s - %(levelname)s - %(message)s"
date_format = "%Y-%m-%d"

logging.basicConfig(level=logging.INFO, format=log_format, datefmt=date_format)


def load(filepath: Path | str) -> dict:
    """Loads the abvio yaml file into a dictionary

    Args:
        filepath: The path to the abvio yaml file

    Returns:
        dict: The dictionary containing vasp input information
    """

    log.debug(f"Loading file: {filepath}")

    if not os.path.exists(filepath):
        log.error(f"File does not exist: {filepath}")
        raise FileNotFoundError(f"File does not exist: {filepath}")

    with open(filepath, "r") as f:
        input_dictionary = yaml.safe_load(f)

    return input_dictionary


def format_structure_output(structure: Structure) -> dict:
    """Formats the structure object into a dictionary

    This is used for the to_file() method in the Input class

    Args:
        structure (Structure): The pymatgen structure object

    Returns:
        dict: The dictionary representation of the structure
    """

    #structure dict only needs lattice: a,b,c, alpha, beta, gamma ; species: list of species ; coords: list of coordinates
    structure_dict = {
        "mode": "manual",
        "lattice": {
            "a": structure.lattice.a,
            "b": structure.lattice.b,
            "c": structure.lattice.c,
            "alpha": structure.lattice.alpha,
            "beta": structure.lattice.beta,
            "gamma": structure.lattice.gamma,
        },
        "species": [ str(specie) for specie in structure.species ],
        "coords": structure.cart_coords.tolist(),
    }

    return structure_dict


def conver_pmg_kpoints_name_to_abvio_name(pmg_name: str) -> str:
    """Converts a pymatgen kpoints name to an abvio kpoints name

    Args:
        pmg_name (str): The pymatgen kpoints name

    Returns:
        str: The abvio kpoints name
    """

    kpoints_name_map = {
        "Gamma": "gamma",
        "Monkhorst": "monkhorst",
        "Line_mode": "line",
        "Automatic": "surface",
    }

    return kpoints_name_map[str(pmg_name)]


def format_kpoints_output(kpoints: Kpoints) -> dict:
    """Formats the kpoints object into a dictionary

    This is used for the to_file() method in the Input class

    Args:
        kpoints (Kpoints): The pymatgen kpoints object

    Returns:
        dict: The dictionary representation of the kpoints
    """


    kpoints_dict = {
        "mode": conver_pmg_kpoints_name_to_abvio_name(kpoints.style),
        "shift": [float(shift) for shift in kpoints.kpts_shift],
    }

    modes = Kpoints.supported_modes
    simple_list_styles = [modes.Monkhorst, modes.Gamma, modes.Automatic]
    if kpoints.style in simple_list_styles:
        kpoints_dict["spacing"] = list(kpoints.kpts[0])

    elif kpoints.style == Kpoints.supported_modes.Line_mode:
        kpoints_dict["spacing"] = kpoints.num_kpts
        kpoints_dict["paths"] = [list(kpts) for kpts in kpoints.kpts]

    if kpoints.labels is not None:
        kpoints_dict["labels"] = kpoints.labels

    return kpoints_dict



class Input:
    def __init__(self, input_dictionary: dict):
        self.input_dict = input_dictionary
        self.structure_dict = input_dictionary.get("structure")
        self.incar_dict = input_dictionary.get("incar")
        self.kpoints_dict = input_dictionary.get("kpoints")

    @property
    def structure(self) -> Structure:
        if self.structure_dict is None:
            raise ValueError(
                f"No structure dictionary found in input file: keys passed are {self.input_dict.keys()}"
            )

        structure_model = structure_model_from_input_dict(self.structure_dict)
        structure = structure_model.structure

        return structure

    @property
    def incar(self) -> Incar:
        if self.incar_dict is None:
            raise ValueError(
                f"No incar dictionary found in input file: keys passed are {self.input_dict.keys()}"
            )

        incar_model = IncarModel(incar_dict=self.incar_dict)
        incar = incar_model.incar(self.structure)

        return incar

    @property
    def kpoints(self) -> Kpoints:
        if self.kpoints_dict is None:
            raise ValueError(
                f"No kpoints dictionary found in input file: keys passed are {self.input_dict.keys()}"
            )

        kpoints_model = kpoints_model_from_dictionary(self.kpoints_dict)

        if kpoints_model.requires_structure:
            kpoints = kpoints_model.kpoints(self.structure)
        else:
            kpoints = kpoints_model.kpoints()

        return kpoints

    @classmethod
    def from_file(cls, filepath: Path | str):
        """Creates an Input object from a file

        Args:
            filepath: The path to the abvio yaml file

        Returns:
            Input: The Input object
        """

        input_dict = load(filepath)
        return cls(input_dict)

    def write_inputs(self, directory: Path | str):
        """Writes the input files to the directory

        Args:
            directory: The directory to write the input files to
        """

        # write only the files associated with non-None dictionaries
        if self.structure_dict is not None:
            poscar = Poscar(self.structure)
            poscar.write_file(os.path.join(directory, "POSCAR"))

        if self.incar_dict is not None:
            incar = self.incar

            if incar.check_params():
                warnings.warn(
                    "INCAR file contains unrecognized tags or values", UserWarning
                )
            
            incar.write_file(os.path.join(directory, "INCAR"))

        if self.kpoints_dict is not None:
            kpoints = self.kpoints
            kpoints.write_file(os.path.join(directory, "KPOINTS"))

    @classmethod
    def from_vaspset(cls, directory: Path | str) -> dict:
        """Reads the input files from a directory

        Args:
            directory: The directory to read the input files from
        """

        # read the files from the directory
        poscar = Poscar.from_file(os.path.join(directory, "POSCAR"))
        incar = Incar.from_file(os.path.join(directory, "INCAR"))
        kpoints = Kpoints.from_file(os.path.join(directory, "KPOINTS"))

        # create the input dictionary
        structure_dict = format_structure_output(poscar.structure)
        incar_dict = dict(incar)
        kpoints_dict = format_kpoints_output(kpoints)

        input_dict = {
            "structure": structure_dict,
            "incar": incar_dict,
            "kpoints": kpoints_dict
        }

        return cls(input_dict)

    def write_file(self, filename: Path | str):
        """Writes the input dictionary to a file

        Args:
            filepath: The path to write the abvio yaml file to
        """
    
        #prepare structure and kpoints dictionaries
        structure_dict = format_structure_output(self.structure)
        kpoints_dict = format_kpoints_output(self.kpoints)

        #write the dictionary to a file
        output_dict = {
            "structure": structure_dict,
            "incar": self.incar_dict,
            "kpoints": kpoints_dict
        }

        with open(filename, "w") as f:
            yaml.dump(output_dict, f, default_flow_style=None)



import logging


from pymatgen.core import Structure
from pymatgen.io.vasp import Kpoints
from pymatgen.symmetry.bandstructure import HighSymmKpath
from pydantic import BaseModel, field_validator
from typing import List, Union, Tuple

log = logging.getLogger(__name__)
log_format = "%(asctime)s - %(levelname)s - %(message)s"
date_format = "%Y-%m-%d"

logging.basicConfig(level=logging.DEBUG, format=log_format, datefmt=date_format)


Number = Union[int, float]
IntArray3D = Tuple[int, int, int]
Matrix3D = List[Tuple[Number, Number, Number]]


class BaseKpoints(BaseModel):
    """Represents a base class for Kpoints"""

    mode: str
    spacing: int | float | list

    @property
    def requires_structure(self) -> bool:
        """Checks if the Kpoints object requires a structure

        This is used to inform abvio if we need to pass a Structure object to the kpoints method

        Returns:
            bool: True if the Kpoints mode requires a structure, False otherwise
        """
        return self.mode in {"surface", "autoline"}

    @field_validator("mode")
    def check_mode(cls, mode) -> str:
        """Tries to handle different cases of the mode string and returns the correct mode

        Args:
            mode (str): The user specified mode of the kpoints

        Returns:
            str: The correct mode of the kpoints
        """

        if mode.lower().startswith("g"):
            return "gamma"
        elif mode.lower().startswith("m"):
            return "monkhorst"
        elif mode.lower() == "surface":
            return "surface"
        elif mode.lower().startswith("line"):
            return "line"
        elif mode.lower() == "autoline":
            return "autoline"
        else:
            log.error(f"Invalid mode: {mode}")
            raise ValueError(f"Invalid mode: {mode}")


class GammaKpoints(BaseKpoints):
    """Represents a Gamma centered grid of kpoints

    Args:
        BaseKpoints: A pydantic BaseModel

    Attributes:
        mode (str): The mode of the kpoints
        spacing (IntArray3D): The spacing of the kpoints
        shift (IntArray3D): The shift of the kpoints
    """

    mode: str = "gamma"
    spacing: IntArray3D
    shift: IntArray3D = [0, 0, 0]

    def kpoints(self):
        kpoints = Kpoints.gamma_automatic(self.spacing, self.shift)

        return kpoints


class MonkhorstKpoints(BaseKpoints):
    """Represents a Monkhorst-Pack grid of kpoints

    Args:
        BaseKpoints: A pydantic BaseModel

    Attributes:
        mode (str): The mode of the kpoints
        spacing (IntArray3D): The spacing of the kpoints
        shift (IntArray3D): The shift of the kpoints
    """

    mode: str = "monkhorst"
    spacing: IntArray3D
    shift: IntArray3D = [0, 0, 0]

    def kpoints(self):
        kpoints = Kpoints.monkhorst_automatic(self.spacing, self.shift)

        return kpoints


class SurfaceKpoints(BaseKpoints):
    """Represents kpoints for a fermi surface calculation

    Args:
        BaseKpoints: A pydantic BaseModel

    Attributes:
        mode (str): The mode of the kpoints
        spacing (float): The spacing of the kpoints
    """

    mode: str = "surface"
    spacing: float

    def kpoints(self, structure: Structure):
        """Generates kpoints for a fermi surface calculation

        Spacing is the grid density scaled by the number of divisions along each reciprocal lattice vector proportional to its length.
        Usually a large number of kpoints are needed for a fermi surface calculation (e.g 1000)

        Args:
            structure (Structure): The structure to generate kpoints for
        """

        kpoints = Kpoints.automatic_density(structure=structure, kppa=self.spacing)

        return kpoints


class LineKpoints(BaseKpoints):
    """Represents a line mode kpoints

    Args:
        BaseKpoints: A pydantic BaseModel

    Attributes:
        mode (str): The mode of the kpoints
        spacing (Number): The spacing of the kpoints
        paths (Matrix3D): Points that form the path in the Brillouin zone where each entry is [kx, ky, kz]
        labels (List[str]): The labels of the kpoints
    """

    mode: str = "line"
    spacing: int
    paths: Matrix3D
    labels: List[str]

    @field_validator("spacing")
    def check_spacing(cls, spacing) -> int:
        """Checks if the spacing is valid

        Args:
            spacing (int): The spacing of the kpoints

        Returns:
            int: The spacing if it is valid

        Raises:
            ValueError: If the spacing is invalid
        """

        if spacing < 1:
            log.error(f"Spacing must be a non-negative integer: {spacing}")
            raise ValueError(
                f"Spacing must be a non-negative integer (i.e divisions per path): {spacing}"
            )

        return spacing

    @field_validator("paths")
    def check_paths(cls, paths) -> Matrix3D:
        """Checks if the paths are valid

        1. the paths must have at least 2 points
        2. the paths must be a 3D matrix
        3. each point must have 3 coordinates

        Args:
            paths (Matrix3D): The paths to check

        Returns:
            Matrix3D: The paths if they are valid
        """
        if len(paths) < 2:
            log.error(f"Invalid paths: {paths}")
            raise ValueError(f"Invalid paths: {paths}")

        for path in paths:
            if len(path) != 3:
                log.error(f"Invalid path: {path}")
                raise ValueError(f"Invalid path: {path}")

        return paths

    @field_validator("labels")
    def check_labels(cls, labels) -> List[str]:
        """Checks if the labels are valid

        1. labels must be greater than 1
        2. each label must be a string

        Args:
            labels (List[str]): The labels to check

        Returns:
            List[str]: The labels if they are valid

        Raises:
            ValueError: If the labels are invalid
        """

        if len(labels) < 2:
            log.error(f"Number of labels must be greater than 1: {labels}")
            raise ValueError(f"Number of labels must be greater than 1: {labels}")

        for label in labels:
            if not isinstance(label, str):
                log.error(f"Invalid label: {label}")
                raise ValueError(f"Invalid label: {label}")

        return labels

    def kpoints(self, system: str = "reciprocal"):
        """Generates a line mode Kpoints object

        Args:
            system (str): The coordinate system of the path. Defaults to "reciprocal"

        Returns:
            Kpoints: A pymatgen Kpoints object
        """

        if len(self.paths) != len(self.labels):
            log.error(f"Invalid paths and labels: {self.paths}, {self.labels}")
            raise ValueError(
                f"Dimensions of labels and paths do not match: {self.paths}, {self.labels}"
            )

        header = "\n".join(
            ["Line mode KPOINTS file", f"{self.spacing}", "linemode", f"{system}"]
        )
        path_string = "\n".join(
            [
                f"{path[0]} {path[1]} {path[2]} {label}"
                for path, label in zip(self.paths, self.labels)
            ]
        )
        kpoints_str = f"{header}\n{path_string}"

        log.debug(f"Kpoints string: {kpoints_str}")

        kpoints = Kpoints.from_str(kpoints_str)

        return kpoints


class AutoLineKpoints(BaseKpoints):
    mode: str = "autoline"
    spacing: int

    @field_validator("spacing")
    def check_spacing(cls, spacing) -> int:
        """Checks if the spacing is valid

        Args:
            spacing (int): The spacing of the kpoints

        Returns:
            int: The spacing if it is valid

        Raises:
            ValueError: If the spacing is invalid
        """

        if spacing < 1:
            log.error(f"Spacing must be a non-negative integer: {spacing}")
            raise ValueError(
                f"Spacing must be a non-negative integer (i.e divisions per path): {spacing}"
            )

        return spacing

    def kpoints(self, structure: Structure):
        """Automatically generates line mode kpoints

        Args:
            structure (Structure): The structure to generate kpoints for

        Returns:
            Kpoints: A pymatgen Kpoints object
        """

        kpath = HighSymmKpath(structure)
        kpoints = Kpoints.automatic_linemode(self.spacing, kpath)

        return kpoints


def kpoints_model_from_dictionary(kpoints_dictionary: dict) -> Kpoints:
    """Parses and validates a dictionary to create a Kpoints object according to the mode provided"""

    base_model = BaseKpoints.validate(kpoints_dictionary)

    mode_model_map = {
        "gamma": GammaKpoints,
        "monkhorst": MonkhorstKpoints,
        "surface": SurfaceKpoints,
        "line": LineKpoints,
        "autoline": AutoLineKpoints,
    }

    BaseKpointsModel = mode_model_map.get(base_model.mode)
    KpointsModel = BaseKpointsModel.validate(kpoints_dictionary)

    return KpointsModel


def kpoints_from_dictionary(
    kpoints_dictionary, structure: Structure | None = None
) -> Kpoints:
    """Parses and validates a dictionary to create a Kpoints object according to the mode provided"""

    KpointsModel = kpoints_model_from_dictionary(kpoints_dictionary)

    if KpointsModel.requires_structure:
        if structure is None:
            log.error(
                f"Kpoints model requires a structure but give type {type(structure)}"
            )
            raise ValueError("Kpoints model requires a structure")

        kpoints = KpointsModel.kpoints(structure)

    else:
        kpoints = KpointsModel.kpoints()

    return kpoints



import argparse
import sys
import yaml
import logging
import os

from pathlib import Path
from pydantic import ValidationError
from abvio.io import Input
from rich.console import Console
from rich.table import Table

#set logging level to silent
logging.basicConfig(level=logging.ERROR)

def check_input_file(input_file):
    try:
        Input.from_file(input_file)
    except ValueError:
        print(f"Input file is incorrectly formatted: {input_file}")
        sys.exit(1)
    except ValidationError as e:
        print(f"Input file contains incorrect tags or parameters: {input_file}")
        print(e)
        sys.exit(1)
    except yaml.YAMLError as e:
        print(f"Input file is not valid yaml: {input_file}")
        print(e)
        sys.exit(1)
    except FileNotFoundError:
        print(f"Input file not found: {input_file}")
        sys.exit(1)


def preview(input_file):
    """Uses rich to create a preview of the structure, kpoints, and incar files.

    Args:
        input_file (str): The path to the abvio yaml file.

    Returns:
        rich.table.Table: A table with the structure, kpoints, and incar files.
    """
    table = Table(title="VASP Input Preview")

    input_object = Input.from_file(input_file)
    structure = input_object.structure
    incar = input_object.incar
    kpoints = input_object.kpoints

    table.add_column("Structure")
    table.add_column("INCAR")
    table.add_column("KPOINTS")
    table.add_row(f"{structure}", f"{incar}", f"{kpoints}") 

    return table



def main():
    """
    Create a VASP input set from a YAML file.

    This function parses command line arguments, reads an abvio YAML file, and generates VASP input files based on the
    provided input. It also performs optional checks on the input file for validity.

    Command line arguments:
        input (str | Path): The path to the abvio YAML file or directory.
        -o, --output (str): The path to the output directory or file.
        --check: Perform validity checks on the input file.
        --preview: Preview the input files.
        --convert: Convert VASP input files to abvio format.
        --verbose: Increase output verbosity.
    """
    parser = argparse.ArgumentParser(
        description="Create a VASP input set from a YAML file"
    )
    parser.add_argument("input", type=str, help="The path to the abvio YAML file or directory")
    parser.add_argument(
        "-o", "--output", type=str, help="The path to the output directory or file"
    )
    parser.add_argument(
        "--check", action="store_true", help="Check validity of the input file"
    )
    parser.add_argument(
        "--preview", action="store_true", help="Preview the input files"
    )
    parser.add_argument(
        "--convert", action="store_true", help="Convert VASP input files to abvio format"
    )
    parser.add_argument(
        "--verbose", action="store_true", help="Increase output verbosity"
    )
    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)

    user_input = args.input
    output_path = args.output

    if args.check:
        check_input_file(user_input)

    if args.preview:
        Console().print(preview(user_input), markup=False)

    if args.output and not args.convert:
        if not os.path.exists(output_path):
            logging.error(f"Output path '{output_path}' does not exist.")
            sys.exit(1)
        input_object = Input.from_file(user_input)
        input_object.write_inputs(output_path)

    if args.convert:

        input_object = Input.from_vaspset(user_input)
        input_object.write_file(output_path)



if __name__ == "__main__":
    main()



from pymatgen.io.vasp import Poscar
from pymatgen.core import Structure, Lattice
from pydantic import BaseModel, field_validator, model_validator
from typing import List, Union, Tuple
from pathlib import Path

import numpy as np
import logging
import os

log = logging.getLogger(__name__)
log_format = "%(asctime)s - %(levelname)s - %(message)s"
date_format = "%Y-%m-%d"

logging.basicConfig(level=logging.INFO, format=log_format, datefmt=date_format)

manual_set = ["lattice", "species", "coords"]
external_set = ["file", "string", "code"]
prototype_set = ["species", "prototype"]

Number = Union[int, float]
ArrayLike = List[Tuple[Number, Number, Number]]


def is_valid_poscar_file(filepath: Path | str) -> bool:
    """Checks if the file is a valid POSCAR file

    Args:
        filepath: The path to the file

    Returns:
        bool: True if file is a valid POSCAR file
    """

    log.debug(f"Checking if file is a valid POSCAR file: {filepath}")

    try:
        Poscar.from_file(filepath)
        return True
    except Exception:
        return False


def is_valid_poscar_string(poscar_string: str) -> bool:
    """Checks if the string is a valid POSCAR string

    Args:
        poscar_string: The POSCAR string

    Returns:
        bool: True if string is a valid POSCAR string
    """

    log.debug("Checking if string is a valid POSCAR string")

    try:
        Poscar.from_str(poscar_string)
        return True
    except Exception as e:
        log.error(f"Invalid POSCAR string: {e}")
        return False


def format_species(species: list[dict] | list[str]) -> list[str]:
    """Formats the species list

    Species list can either be
    - A list of dictionaries where each dictionary contains the species and count
    - A list of strings where each string is the species

    Args:
        species: The list of species

    Returns:
        list: The formatted list of species
    """

    log.debug("Formatting species list")
    log.debug(f"Species is of type: {type(species)}")

    if not isinstance(species, list):
        log.error(f"Species is of type: {type(species)}")
        raise ValueError("Species must be a list")

    if isinstance(species[0], dict):
        # check if keys are strings and values are integers
        for species_dict in species:
            if not all(isinstance(key, str) for key in species_dict.keys()):
                log.error(
                    f"Keys in species dictionary are not strings: {species_dict.keys()}"
                )
                raise ValueError("Keys in species dictionary must be strings")

            if not all(isinstance(value, int) for value in species_dict.values()):
                log.error(
                    f"Values in species dictionary are not integers: {species_dict.values()}"
                )
                raise ValueError("Values in species dictionary must be integers")

    if isinstance(species[0], dict):
        formatted_species_list = []
        for species_dict in species:
            species_str, species_count = next(iter(species_dict.items()))
            formatted_species_list.extend([species_str] * species_count)

        return formatted_species_list

    if isinstance(species, list) and isinstance(species[0], str):
        return species

    else:
        log.error(f"Species is of type: {type(species)}")
        raise ValueError("Species must be a list of dictionaries or a list of strings")


def structure_from_mpi_code(mpcode: str, is_conventional: bool = True):
    """
    Creates a pymatgen structure from a code
    """
    from mp_api.client import MPRester

    api_key = os.getenv("MP_API_KEY")

    with MPRester(api_key, mute_progress_bars=True) as mpr:
        log.debug("Connected to materials project API")
        structure = mpr.get_structure_by_material_id(
            mpcode, conventional_unit_cell=is_conventional
        )

    if structure is None:
        raise ValueError(f"Could not find structure with code {mpcode}")

    return structure


def structure_from_lattice(
    lattice: Lattice | ArrayLike,
    species: list[str],
    coords: ArrayLike,
    cartesian: bool = True,
) -> Structure:
    """Create a Poscar object from a lattice object, species list, and coordinates array"""

    log_data = {
        "Lattice": lattice,
        "Species": species,
        "Coordinates": coords,
        "Cartesian": cartesian,
    }

    log.debug(f"Creating structure from lattice array: {log_data}")

    if isinstance(lattice, list) or isinstance(lattice, np.ndarray):
        lattice = Lattice(lattice)

    if not isinstance(lattice, Lattice):
        log.error(f"Expected array or Lattice object, got {type(lattice)}")
        raise ValueError("Lattice must be a Lattice object or array")

    if len(coords) != len(species):
        log.error(
            f"Species  ({len(species)}) and coordinates ({len(coords)}) are not the same length"
        )
        raise ValueError(
            f"Number of species {len(species)} and coordinates {len(coords)} must be the same"
        )

    structure = Structure(lattice, species, coords, coords_are_cartesian=cartesian)

    return structure


def structure_from_prototype(
    prototype: str, lattice: dict, species: list[str], cartesian: bool = True
) -> Structure:
    """Create a structure from a prototype, lattice object, species list, and coordinates array"""

    log_dict = {
        "Prototype": prototype,
        "Species": species,
        "Lattice": lattice,
        "Cartesian": cartesian,
    }

    log.debug(f"Creating structure from prototype: {log_dict}")

    if isinstance(lattice, Lattice):
        lattice = lattice.as_dict()

    structure = Structure.from_prototype(prototype, species, **lattice)

    return structure


def structure_from_file(file: Path | str) -> Structure:
    """Creates a pymatgen Structure object from a file containing structure information

    Args:

        file (Path | str): The path to the file containing the structure information

    Returns:

            Structure: The pymatgen Structure object created from the file
    """

    log.debug(f"Creating structure from file: {file}")

    return Structure.from_file(file)


def structure_from_string(string: str) -> Structure:
    """Creates a pymatgen Structure object from a string containing structure information

    Args:

        string (str): The string containing the structure information

    Returns:

            Structure: The pymatgen Structure object created from the string
    """

    log.debug(f"Creating structure from string: {string}")

    return Poscar.from_str(string).structure


class BaseStructure(BaseModel):
    mode: str

    @field_validator("mode")
    def check_mode(cls, mode) -> str:
        """Tries to handle different cases of the mode string and returns the correct mode

        Args:
            mode (str): The user specified mode of the kpoints

        Returns:
            str: The correct mode of the kpoints
        """

        if mode.lower().startswith("ext"):
            return "external"
        elif mode.lower().startswith("pro"):
            return "prototype"
        elif mode.lower() == "manual":
            return "manual"
        else:
            log.error(f"Invalid mode: {mode}")
            raise ValueError(f"Invalid mode: {mode}")


class ManualStructure(BaseStructure):
    """Handles validating manual structure input

    Requires lattice, species, and coords
    """

    mode: str = "manual"
    lattice: Lattice | ArrayLike | dict
    coords: ArrayLike
    species: list[str] | list[dict[str, int]]

    @field_validator("species")
    def validate_species(cls, species):
        return format_species(species)

    @field_validator("lattice")
    def validate_lattice(cls, lattice):
        if isinstance(lattice, list) or isinstance(lattice, np.ndarray):
            lattice = Lattice(lattice)
        elif isinstance(lattice, dict):
            lattice = Lattice.from_dict(lattice)
        return lattice

    @property
    def structure(self) -> Structure:
        """Creates a pymatgen Structure object from the input data"""
        return structure_from_lattice(self.lattice, self.species, self.coords)


class PrototypeStructure(BaseStructure):
    """Pydantic model for prototype structure input

    Must contain species and prototype

    Valid prototypes:
    fcc, bcc, hcp, diamond, rocksalt, perovskite, cscl, fluorite, antifluorite, zincblende
    """

    mode: str = "prototype"
    species: list[str] | list[dict]
    lattice: dict
    prototype: str

    @field_validator("species")
    def validate_species(cls, species):
        return format_species(species)

    @field_validator("prototype")
    def validate_prototype(cls, prototype):
        valid_prototypes = [
            "fcc",
            "bcc",
            "hcp",
            "diamond",
            "rocksalt",
            "perovskite",
            "cscl",
            "fluorite",
            "antifluorite",
            "zincblende",
        ]
        if prototype not in valid_prototypes:
            raise ValueError(f"Prototype '{prototype}' is not implemented")
        return prototype

    @property
    def structure(self) -> Structure:
        """Creates a pymatgen Structure object from the input data"""

        return structure_from_prototype(
            prototype=self.prototype, species=self.species, lattice=self.lattice
        )


class ExternalStructure(BaseStructure):
    """Pydantic model for external structure input

    Must contain either file, string, or code
    """

    mode: str = "external"
    file: Path | str = None
    string: str = None
    code: str = None

    @model_validator(mode="before")
    def check_external(cls, values):
        if not any(values.values()):
            raise ValueError("Must provide either file, string, or code")

        # if sum(bool(values[key]) for key in values) > 1:
        #     raise ValueError("Must provide only one of file, string, or code")
        return values

    @field_validator("file")
    def check_file(cls, file):
        if not is_valid_poscar_file(file):
            raise ValueError(f"Invalid POSCAR file: {file}")
        return file

    @field_validator("string")
    def check_string(cls, string):
        if not is_valid_poscar_string(string):
            raise ValueError("Invalid POSCAR string")
        return string

    @field_validator("code")
    def check_code(cls, code):
        # code is a string that starts with "mp-" and is followed by a number
        if not code.startswith("mp-") or not code[3:].isdigit():
            raise ValueError("Invalid Materials Project code")
        return code

    @property
    def structure(self) -> Structure:
        """Checks to see if file, string, or code is provided and returns a pymatgen Structure object"""

        if self.file and self.string and self.code:
            log.debug(
                f"was given file: {self.file}, string: {self.string}, and code: {self.code}"
            )
            raise ValueError("Must provide only one of file, string, or code")

        elif self.file:
            struct = structure_from_file(self.file)

        elif self.string:
            struct = structure_from_string(self.string)

        elif self.code:
            struct = structure_from_mpi_code(self.code)

        else:
            raise ValueError("Must provide either file, string, or code")

        return struct


def structure_model_from_input_dict(structure_dictionary: dict) -> Structure:
    """Creates a pymatgen Structure object from a dictionary containing input structure information

    Args:

        structure_dict (dict): The dictionary containing the structure information

    Returns:

            Structure: The pymatgen Structure object created from the dictionary
    """

    log.debug(f"Creating structure from input dictionary: {structure_dictionary}")

    base_model = BaseStructure.validate(structure_dictionary)

    mode_model_map = {
        "manual": ManualStructure,
        "prototype": PrototypeStructure,
        "external": ExternalStructure,
    }

    BaseStructureModel = mode_model_map.get(base_model.mode)
    StructureModel = BaseStructureModel.validate(structure_dictionary)

    return StructureModel



import unittest
import yaml
import os
import abvio.io as io
import numpy as np

from pathlib import Path
from pymatgen.core import Structure
from pymatgen.io.vasp import Incar, Kpoints, Poscar

from pymatgen.symmetry.analyzer import SpacegroupAnalyzer


base_path = Path(__file__).parent
structure_dir = os.path.join(base_path, 'structures')
files_dir = os.path.join(base_path, 'files')
vaspset_dir = os.path.join(base_path, 'vaspsets')



class TestLoadFunction(unittest.TestCase):
    """Contains tests that check if read can load files correctly"""

    def test_load_invalid_file(self):
        """Test if load can read an invalid yaml file"""

        invalid_file = os.path.join(files_dir, 'invalid.yaml')

        with self.assertRaises(yaml.scanner.ScannerError):
            io.load(invalid_file)

    def test_load_invalid_yaml(self):
        """Test if load can read an invalid yaml file"""

        invalid_file = os.path.join(files_dir, 'invalid.yaml')

        with self.assertRaises(yaml.YAMLError):
            io.load(invalid_file)
    

class TestCreateInput(unittest.TestCase):
    """Tests if the input class can create a valid VASP input set"""

    def test_Input_from_file(self):
        """Tests if the input class can create a valid VASP input set"""

        InputObject = io.Input.from_file(os.path.join(files_dir, 'valid.yaml'))
        self.assertIsInstance(InputObject, io.Input)

    def test_kpoints_method(self):
        """Tests if the kpoints method returns a Kpoints object"""

        InputObject = io.Input.from_file(os.path.join(files_dir, 'valid.yaml'))
        kpoints = InputObject.kpoints
        self.assertIsInstance(kpoints, Kpoints)

        for kpt in kpoints.kpts:
            self.assertEqual(kpt, (5, 5, 5))

    def test_incar_method(self):
        """Tests if the incar method returns an Incar object
        
        Incar section looks like the following 
        incar:
            ediff: 1e-6
            ediffg: -0.01
            isif: 3
            nsw: 4
            ibrion: 2
            magmom:
                Ca: 2.0
                F: 0.6
        """

        InputObject = io.Input.from_file(os.path.join(files_dir, 'valid.yaml'))
        incar = InputObject.incar
        self.assertIsInstance(incar, Incar)

        self.assertEqual(float(incar['EDIFF']), 1e-6)
        self.assertEqual(incar['EDIFFG'], -0.01)
        self.assertEqual(incar['ISIF'], 3)
        self.assertEqual(incar['NSW'], 4)
        self.assertEqual(incar['IBRION'], 2)
        self.assertEqual(incar['MAGMOM'], [2] * 4 + [0.6] * 8)

    
    def test_structure_method(self):
        """Tests if the structure method returns a Structure object"""

        InputObject = io.Input.from_file(os.path.join(files_dir, 'valid.yaml'))
        structure = InputObject.structure
        self.assertIsInstance(structure, Structure)

    def test_write_method(self):
        """Tests if the write method writes the input files to the directory"""

        InputObject = io.Input.from_file(os.path.join(files_dir, 'valid.yaml'))
        InputObject.write_inputs('/tmp')

        self.assertTrue(os.path.exists('/tmp/POSCAR'))
        self.assertTrue(os.path.exists('/tmp/INCAR'))
        self.assertTrue(os.path.exists('/tmp/KPOINTS'))

        #check if all files are valid
        Poscar.from_file('/tmp/POSCAR')
        Incar.from_file('/tmp/INCAR')
        Kpoints.from_file('/tmp/KPOINTS')

        os.remove('/tmp/POSCAR')
        os.remove('/tmp/INCAR')
        os.remove('/tmp/KPOINTS')


class TestSiBandStructure(unittest.TestCase):
    """Tests if the Si band structure is generated correctly"""

    def test_incar(self):
        """Tests if the INCAR file is generated correctly"""

        si_band_dir = os.path.join(vaspset_dir, 'band')
        input_file = os.path.join(si_band_dir, 'equivalent.yaml')
        InputObject = io.Input.from_file(input_file)
        
        expected_incar = Incar.from_file(os.path.join(si_band_dir, 'INCAR'))

        self.assertEqual(InputObject.incar, expected_incar)

    def test_kpoints(self):
        """Tests if the KPOINTS file is generated correctly"""

        si_band_dir = os.path.join(vaspset_dir, 'band')
        input_file = os.path.join(si_band_dir, 'equivalent.yaml')
        InputObject = io.Input.from_file(input_file)
        
        expected_kpoints = Kpoints.from_file(os.path.join(si_band_dir, 'KPOINTS'))

        self.assertEqual(set(InputObject.kpoints.kpts), set(expected_kpoints.kpts))

    def test_structure(self):
        """Tests if the structure is generated correctly"""

        si_band_dir = os.path.join(vaspset_dir, 'band')
        input_file = os.path.join(si_band_dir, 'equivalent.yaml')
        InputObject = io.Input.from_file(input_file)
        
        expected_structure = Poscar.from_file(os.path.join(si_band_dir, 'POSCAR')).structure

        self.assertEqual(InputObject.structure.reduced_formula, expected_structure.reduced_formula)

    
class TestPerovskiteSet(unittest.TestCase):
    """Tests if the Perovskite set is generated correctly"""

    def test_incar(self):
        """Tests if the INCAR file is generated correctly"""

        perovskite_dir = os.path.join(vaspset_dir, 'perovskite')
        input_file = os.path.join(perovskite_dir, 'equivalent.yaml')
        InputObject = io.Input.from_file(input_file)
        
        expected_incar = Incar.from_file(os.path.join(perovskite_dir, 'INCAR'))

        #check only magmom, lwave, and encut
        self.assertEqual(InputObject.incar['MAGMOM'], expected_incar['MAGMOM'])
        self.assertEqual(InputObject.incar['LWAVE'], expected_incar['LWAVE'])
        self.assertEqual(InputObject.incar['ENCUT'], expected_incar['ENCUT'])

    def test_kpoints(self):
        """Tests if the KPOINTS file is generated correctly"""

        perovskite_dir = os.path.join(vaspset_dir, 'perovskite')
        input_file = os.path.join(perovskite_dir, 'equivalent.yaml')
        InputObject = io.Input.from_file(input_file)
        
        expected_kpoints = Kpoints.from_file(os.path.join(perovskite_dir, 'KPOINTS'))

        self.assertEqual(set(InputObject.kpoints.kpts), set(expected_kpoints.kpts))

    def test_structure(self):
        """Tests if the structure is generated correctly"""

        perovskite_dir = os.path.join(vaspset_dir, 'perovskite')
        input_file = os.path.join(perovskite_dir, 'equivalent.yaml')
        InputObject = io.Input.from_file(input_file)
        
        expected_structure = Poscar.from_file(os.path.join(perovskite_dir, 'POSCAR')).structure

        self.assertEqual(InputObject.structure.reduced_formula, expected_structure.reduced_formula)
        



if __name__ == '__main__':
    unittest.main()


import logging

from pymatgen.core import Structure
from pymatgen.io.vasp import Incar

from pydantic import BaseModel, field_validator

log = logging.getLogger(__name__)
log_format = "%(asctime)s - %(levelname)s - %(message)s"
date_format = "%Y-%m-%d"

logging.basicConfig(level=logging.INFO, format=log_format, datefmt=date_format)


def assign_site_property_by_species(
    structure: Structure,
    species: str,
    property_name: str,
    property_value: float | list[float],
) -> Structure:
    """
    Assign a property to a species
    Args:
        structure: pymatgen Structure object
        species: species to assign property to
        property_name: name of property to assign
        property_value: value of property to assign

    Returns:
        pymatgen Structure object with property assigned
    """

    sites = structure.sites
    for site in sites:
        if site.species_string == species:
            site.properties[property_name] = property_value

    return structure


def assign_site_property_by_index(
    structure: Structure,
    index: int,
    property_name: str,
    property_value: float | list[float],
) -> Structure:
    """
    Assign a property to a site
    Args:
        structure: pymatgen Structure object
        index: index of site to assign property to
        property_name: name of property to assign
        property_value: value of property to assign

    Returns:
        pymatgen Structure object with property assigned
    """

    sites = structure.sites
    sites[index].properties[property_name] = property_value

    return structure


def assign_site_property_by_range(
    structure: Structure,
    range: slice,
    property_name: str,
    property_value: float | list[float],
) -> Structure:
    """
    Assign a property to a range of sites
    Args:
        structure: pymatgen Structure object
        range: slice object used to specify range of sites to assign property to
        property_name: name of property to assign
        property_value: value of property to assign

    Returns:
        pymatgen Structure object with property assigned
    """

    sites = structure.sites
    for site in sites[range]:
        site.properties[property_name] = property_value

    return structure


def is_range_dict(range_dict: dict) -> bool:
    """
    Check if a dictionary is a range dictionary

    Format is a dictionary that contains
    {
        "start": int,
        "stop": int,
        "value" : (float | list[float])
    }

    Sometimes the dictionary may contain "step" key to specify step size

    Args:
        range_dict: dictionary to check
    Returns:
        bool indicating if dictionary is a range dictionary
    """

    if "start" in range_dict and "stop" in range_dict and "value" in range_dict:
        return True
    else:
        log.error(f"Range dictionary is improperly formatted: {range_dict}")
        return False


def is_range_list(range_list: list) -> bool:
    """
    Ranges are often specified as a list of dictionaries

    This function checks if a list contains range dictionaries

    Args:
        range_list: list of dictionaries to check
    Returns:
        bool indicating if list contains range dictionaries
    """

    if isinstance(range_list, list) and all(
        [is_range_dict(range_dict) for range_dict in range_list]
    ):
        return True

    log.debug(f"Not a range list: {range_list}")
    return False


def is_species_dict(species_dict: dict) -> bool:
    """
    Check if a dictionary is an index dictionary

    Format is a dictionary that contains
    {
        symbol1: magmom1,
        symbol2: magmom2,
        ...
    }

    Args:
        species_dict: dictionary to check

    Returns:
        bool indicating if dictionary is an index dictionary
    """
    if not isinstance(species_dict, dict):
        return False

    if all(
        [isinstance(value, (int, float, list)) for value in species_dict.values()]
    ) and all([isinstance(key, str) for key in species_dict.keys()]):
        return True
    else:
        return False


def is_index_dict(index_dict: dict) -> bool:
    """
    Check if a dictionary is an index dictionary

    Format is a dictionary that contains
    {
        index1: magmom1,
        index2: magmom2,
        ...
    }

    Args:
        index_dict: dictionary to check

    Returns:
        bool indicating if dictionary is an index dictionary
    """

    if not isinstance(index_dict, dict):
        return False

    if all(
        [isinstance(value, (int, float, list)) for value in index_dict.values()]
    ) and all([isinstance(key, int) for key in index_dict.keys()]):
        return True
    else:
        return False


def is_collinear(magmom_entry: list | dict) -> bool:
    """
    Check if magmom_entry is a collinear magmom entry
    Args:
        magmom_entry: list or dictionary to check
    Returns:
        bool indicating if magmom_entry is a collinear magmom entry
    Raises:
        ValueError if magmom_entry is improperly formatted
    """

    # checks if magmom_entry is an index, range, or species dictionary
    if is_index_dict(magmom_entry) or is_species_dict(magmom_entry):
        # check if values are floats or lists of floats
        if all([isinstance(value, (float, int)) for value in magmom_entry.values()]):
            return True
        if all([isinstance(value, list) for value in magmom_entry.values()]):
            return False
        else:
            raise ValueError(f"Magnetic moments is improperly formatted {magmom_entry}")

    if is_range_list(magmom_entry):
        # select the "value" key of each entry in the list and check if it is a float or list of floats
        if all(
            [
                isinstance(range_dict["value"], (float, int))
                for range_dict in magmom_entry
            ]
        ):
            return True
        if all([isinstance(range_dict["value"], list) for range_dict in magmom_entry]):
            return False
        else:
            raise ValueError(
                f"Magnetic moments is improperly formatted: {magmom_entry}"
            )

    else:
        raise ValueError(f"Magnetic moments is improperly formatted: {magmom_entry}")


def is_valid_magmom_entry(magmom_entry: list | dict) -> bool:
    """
    Check if magmom_entry is a valid magmom entry
    Args:
        magmom_entry: list or dictionary to check
    Returns:
        bool indicating if magmom_entry is a valid magmom entry
    """

    # checks if magmom_entry is an index, range, or species dictionary
    if (
        is_index_dict(magmom_entry)
        or is_species_dict(magmom_entry)
        or is_range_list(magmom_entry)
    ):
        return True
    else:
        return False


def site_properties_from_structure(
    structure: Structure, property_name: str, default=None
) -> list:
    """
    Get a property from a structure
    Args:
        structure: pymatgen Structure object
        property_name: name of property to get
        default: default value of property if no property is assigned
    Returns:
        list of property values
    """

    sites = structure.sites
    properties = [site.properties.get(property_name, default) for site in sites]

    return properties


def format_magnetic_moments(
    magmom_entry: list | dict, structure: Structure, missing: float | list = 0
) -> list:
    """
    Format magnetic moments to be compatible with pymatgen Incar object
    Args:
        magmom_entry: magmom entry to format
        structure: structure to assign magmom entries to
        missing: default value for missing magmom entries
    Returns:
        list of magnetic moments
    """

    if is_index_dict(magmom_entry):
        for index, magmom in magmom_entry.items():
            structure = assign_site_property_by_index(
                structure, index, "magmom", magmom
            )

    if is_species_dict(magmom_entry):
        for species, magmom in magmom_entry.items():
            structure = assign_site_property_by_species(
                structure, species, "magmom", magmom
            )

    if is_range_list(magmom_entry):
        for range_dict in magmom_entry:
            start = range_dict["start"]
            stop = range_dict["stop"]
            step = range_dict.get("step")
            value = range_dict["value"]
            structure = assign_site_property_by_range(
                structure, slice(start, stop, step), "magmom", value
            )

    magmoms = site_properties_from_structure(structure, "magmom", missing)

    return magmoms


def format_incar_dict(
    incar_dict: dict, structure: Structure, missing: float | list = 0
) -> dict:
    """Takes incar dict and formats it to be compatible with pymatgen Incar object

    Currently only supports magmom entries

    Args:
        incar_dict (dict): The incar dictionary
        structure (Structure, optional): The structure to assign magmom entries to
        missing (float | list): The default value for missing magmom entrie. Defaults to 0.
    """
    formatted_incar_dict = incar_dict.copy()

    if formatted_incar_dict.get("magmom"):
        magmom_entry = formatted_incar_dict.pop("magmom")
        magmoms = format_magnetic_moments(magmom_entry, structure, missing)
        formatted_incar_dict["magmom"] = magmoms

    # set all keys to uppercase
    formatted_incar_dict = {
        key.upper(): value for key, value in formatted_incar_dict.items()
    }

    return formatted_incar_dict


class IncarModel(BaseModel):
    """Pydantic model used to help validate the incar dictionary format

    1. checks if incar is a dictionary
    2. checks if all key value pairs are valid
    3. uses pymatgen to validate the incar dictionary
    """

    incar_dict: dict

    @field_validator("incar_dict")
    def validate_incar_dict(cls, incar_dict: dict):
        """Validates the incar dictionary

        Args:
            incar_dict (dict): The incar dictionary

        Raises:
            ValueError: If the incar dictionary is invalid
        """

        if not isinstance(incar_dict, dict):
            raise ValueError("incar_dict must be a dictionary")

        try:
            if "magmom" in incar_dict:
                if not is_valid_magmom_entry(incar_dict["magmom"]):
                    raise ValueError(f"Invalid magmom entry {incar_dict['magmom'] }")
            Incar(incar_dict)
        except Exception as e:
            raise ValueError(f"Invalid INCAR dictionary: {e}")

        return incar_dict

    def incar(self, structure: Structure) -> Incar:
        """Returns the pymatgen Incar object

        Returns:
            Incar: The pymatgen Incar object
        """

        formatted_incar_dictionary = format_incar_dict(self.incar_dict, structure)
        incar = Incar(formatted_incar_dictionary)

        return incar



import abvio.kpoints as kp 

import unittest
import os
from pathlib import Path

from pymatgen.io.vasp import Kpoints
from pymatgen.core import Structure
from pydantic import ValidationError


base_path = Path(__file__).parent
structure_dir = os.path.join(base_path, 'structures')
files_dir = os.path.join(base_path, 'files')
vaspset_dir = os.path.join(base_path, 'vaspsets')

class TestKpointsBaseModel(unittest.TestCase):

    def test_base_kpoints_valid(self):
        kpoints = kp.BaseKpoints(mode="surface", spacing=1)
        self.assertTrue(kpoints.requires_structure)

        kpoints = kp.BaseKpoints(mode="autoline", spacing=1)
        self.assertTrue(kpoints.requires_structure)

    
    def test_variable_mesh_enums(self): 

        gamma_test_cases = ['gamma', 'Gamma', 'GAMMA', 'g', 'G']
        monkhorst_test_cases = ['monkhorst', 'Monkhorst', 'MONKHORST', 'm', 'M']

        for case in gamma_test_cases:
            kpoints = kp.BaseKpoints(mode=case, spacing=1)
            self.assertEqual(kpoints.mode, 'gamma')

        for case in monkhorst_test_cases:
            kpoints = kp.BaseKpoints(mode=case, spacing=1)
            self.assertEqual(kpoints.mode, 'monkhorst')


    def test_base_kpoints_invalid(self):
        with self.assertRaises(ValueError):
            kp.BaseKpoints(mode=9, spacing=1, shift=(0, 0, 0))

        with self.assertRaises(ValueError):
            kp.BaseKpoints(mode="wrong", spacing=1, shift=(0, 0, 0))


class TestLineModeKpoints(unittest.TestCase):
    """Tests if we can create a LineModeKpoints object correctly"""

    def test_valid_linemode(self):
        model = kp.LineKpoints(spacing=30, paths=[[0, 0, 0], [0.5, 0.5, 0.5]], labels=["G", "X"])
        self.assertFalse(model.requires_structure)
        
        kpoints = model.kpoints()
        self.assertIsInstance(kpoints, Kpoints)

    def test_invalid_linemode(self):
        
        with self.assertRaises(ValueError):
            model = kp.LineKpoints(spacing=30, paths=[[0, 0], [0.5, 0.5]], labels=["G", "X"])

        with self.assertRaises(ValueError):
            model = kp.LineKpoints(spacing=30, paths=[[0, 0, 0 ], [0.5, 0.5, 0.5]], labels=["G"])

        with self.assertRaises(ValueError):
            model = kp.LineKpoints(spacing=30, paths=[[0, 0, 0], [0.5, 0.5, 0.5]], labels=["G", "X", "Y"])
            model.kpoints()

        with self.assertRaises(ValueError):
            model = kp.LineKpoints(spacing=0.1, paths=[[0, 0, 0], [0.5, 0.5, 0.5]], labels=["G", "X"])
            model.kpoints()

class TestAutoLineKpoints(unittest.TestCase):
    """Tests if we can create an AutoLineKpoints object correctly"""

    def test_fluorite_autoline(self):
        kpoints_divisions = 10
        model = kp.AutoLineKpoints(spacing=kpoints_divisions)
        self.assertTrue(model.requires_structure)

        structure = Structure.from_file(os.path.join(structure_dir, 'CaF2.vasp'))
        kpoints = model.kpoints(structure)
        self.assertIsInstance(kpoints, Kpoints)
        self.assertEqual(kpoints.num_kpts, kpoints_divisions)

    def test_perovskite_auto_line(self):
        kpoints_divisions = 35
        model = kp.AutoLineKpoints(spacing=kpoints_divisions)
        structure = Structure.from_file(os.path.join(structure_dir, 'CaTiO3.vasp'))
        kpoints = model.kpoints(structure)
        self.assertIsInstance(kpoints, Kpoints)
        self.assertEqual(kpoints.num_kpts, kpoints_divisions)

    def test_invalid_autoline(self):
        with self.assertRaises(ValueError):
            model = kp.AutoLineKpoints(spacing=0.1)
            model.kpoints(None)


class TestSurfaceKpoints(unittest.TestCase):

    def test_valid_surface(self):
        model = kp.SurfaceKpoints(spacing=80000)
        structure = Structure.from_file(os.path.join(structure_dir, 'CaF2.vasp'))
        kpoints = model.kpoints(structure)
    
        self.assertTrue(model.requires_structure)
        self.assertIsInstance(kpoints, Kpoints)

class TestGammaKpoints(unittest.TestCase):

    def test_valid_gamma(self):
        model = kp.GammaKpoints(spacing=[3, 3, 3])
        self.assertFalse(model.requires_structure)
        kpoints = model.kpoints()
        self.assertIsInstance(kpoints, Kpoints)

    def test_compare_from_file(self):
        model = kp.GammaKpoints(spacing=[7, 7, 7])
        kpoints = [list(kpt) for kpt in  model.kpoints().kpts ]
        kpoints_from_file = Kpoints.from_file(os.path.join(vaspset_dir, 'fluorite', 'KPOINTS'))
        expected_kpoints = [list(kpt) for kpt in kpoints_from_file.kpts]
        self.assertEqual(kpoints, expected_kpoints)
    
    def test_write_kpoints(self):
        model = kp.GammaKpoints(spacing=[7, 7, 7])
        kpoints = model.kpoints()
        kpoints.write_file(os.path.join('/tmp', 'test.kpoints'))


class TestMonkhorstKpoints(unittest.TestCase):

    def test_valid_monkhorst(self):
        input_spacing = [2, 2, 1]
        model = kp.MonkhorstKpoints(spacing=input_spacing)
        kpoints = model.kpoints()
        self.assertIsInstance(kpoints, Kpoints)

        for kpoint,expected_value in zip(kpoints.kpts[0], input_spacing):
            self.assertEqual(kpoint, expected_value)
                    

class TestAutoLinemodeFromInputDictionary(unittest.TestCase):

    def test_automatic_linemode_kpoints(self):
        """This test mimics the following yaml input

        kpoints:
            mode: autoline
            spacing: 20

        """
        manual_model = kp.AutoLineKpoints(spacing=20)

        test_dict = {
            'kpoints': {'mode': 'autoline', 'spacing': 20}
        }
        test_model = kp.AutoLineKpoints.validate(test_dict['kpoints'])

        self.assertEqual(test_model, manual_model)

    def test_automatic_linemode_kpoint_gen(self):
        """Test creating Kpoints object from the input dictionary"""

        test_dict = {
            'kpoints': {'mode': 'autoline', 'spacing': 20}
        }
        test_model = kp.AutoLineKpoints.validate(test_dict['kpoints'])

        structure = Structure.from_file(os.path.join(structure_dir, 'CaF2.vasp'))
        kpoints = test_model.kpoints(structure)

        self.assertIsInstance(kpoints, Kpoints)

    def test_invalid_autoline_kpoints(self):
        test_dict = {
            'kpoints': {'mode': 'autoline', 'spacing': 0.1}
        }

        with self.assertRaises(ValidationError):
            kp.AutoLineKpoints.validate(test_dict['kpoints'])

    def test_surface_kpoints(self):
        test_dict = {
            'kpoints': {'mode': 'surface', 'spacing': 80000}
        }
        test_model = kp.SurfaceKpoints.validate(test_dict['kpoints'])
        structure = Structure.from_file(os.path.join(structure_dir, 'CaF2.vasp'))
        kpoints = test_model.kpoints(structure)

        self.assertIsInstance(test_model, kp.SurfaceKpoints)
        self.assertIsInstance(kpoints, Kpoints)

    
class TestLinemodeFromInputDictionary(unittest.TestCase):
    
    def test_linemode_kpoints(self):
        """This test mimics the following yaml input
    
        kpoints:
            mode: line
            spacing: 30
            paths: [[0, 0, 0], [0.5, 0.5, 0.5]]
            labels: ["G", "X"]
    
        """
        manual_model = kp.LineKpoints(spacing=30, paths=[[0, 0, 0], [0.5, 0.5, 0.5]], labels=["G", "X"])
    
        test_dict = {
            'kpoints': {'mode': 'line', 'spacing': 30, 'paths': [[0, 0, 0], [0.5, 0.5, 0.5]], 'labels': ["G", "X"]}
        }
        test_model = kp.LineKpoints.validate(test_dict['kpoints'])
    
        self.assertEqual(test_model, manual_model)
    
    def test_linemode_kpoint_gen(self):
        """Test creating Kpoints object from the input dictionary"""
    
        test_dict = {
            'kpoints': {'mode': 'line', 'spacing': 30, 'paths': [[0, 0, 0], [0.5, 0.5, 0.5]], 'labels': ["G", "X"]}
        }
        test_model = kp.LineKpoints.validate(test_dict['kpoints'])
        kpoints = test_model.kpoints()
    
        self.assertIsInstance(kpoints, Kpoints)
    
    def test_invalid_linemode_kpoints(self):
        test_dict = {
            'kpoints': {'mode': 'line', 'spacing': 30, 'paths': [[0, 0], [0.5, 0.5]], 'labels': ["G", "X"]}
        }
    
        with self.assertRaises(ValidationError):
            kp.LineKpoints.validate(test_dict['kpoints'])
    
    def test_invalid_linemode_kpoints_labels(self):
        test_dict = {
            'kpoints': {'mode': 'line', 'spacing': 30, 'paths': [[0, 0, 0], [0.5, 0.5, 0.5]], 'labels': ["G"]}
        }
    
        with self.assertRaises(ValidationError):
            kp.LineKpoints.validate(test_dict['kpoints'])
    

class TestKpointsModeDetection(unittest.TestCase):

    valid_test_dicts = [
        {
            'kpoints': {'mode': 'line', 'spacing': 30, 'paths': [[0, 0, 0], [0.5, 0.5, 0.5]], 'labels': ["G", "X"]}
        },
        {
            'kpoints': {'mode': 'autoline', 'spacing': 20}
        },
        {
            'kpoints': {'mode': 'surface', 'spacing': 80000}
        },
        {
            'kpoints': {'mode': 'gamma', 'spacing': [3, 3, 3]}
        },
        {
            'kpoints': {'mode': 'Monkhorst', 'spacing': [2, 2, 1]}
        },
        {
            'kpoints': {'mode': 'monkhorst-pack', 'spacing': [4, 2, 7]}
        }
    ]

    def test_valid_kpoints_mode_detection(self):
        test_structure = Structure.from_file(os.path.join(structure_dir, 'CaF2.vasp'))
        for test_dict in self.valid_test_dicts:
            kpoints = kp.kpoints_from_dictionary(test_dict['kpoints'], structure=test_structure)
            self.assertIsInstance(kpoints, Kpoints)
            



if __name__ == "__main__":
    unittest.main()